/****************************************************************************
*  Browse Function By: Tom Claffy  March 1993
*  Search Routine  By: Phil Barnett  April 3, 1993
*  Minor Repair of DownFillArray By: Phil Barnett  April 5, 1993
*  Added Highlight to Search By: Tom Claffy  April 10, 1993
*  Added minor comments and fixed anomalous display on small files
*                                Tom Claffy  May 25, 1993
*  Added MAXBROWSELENGTH to avoid lockup on binary files: TC 7-26-1993
*  Fixed Ctrl-PgUp and Ctrl-PgDn problems with small files: TC 9-15-93
*  Added optional code block to modify text as it is extracted TC 9-23-93
*  LineDisp Function By: Todd C. MacDonald 9-28-1993 with minor
*      modification by Tom Claffy
*  Minor fixes and additional documentation: Tom Claffy April 1994
*  Help screen: Bryan Smith November 1994
*
*  Placed in Public Domain July 29, 1993 by Tom Claffy
*
*  Pure Clipper Text File Browser
*
*  with...
*
*      Virtual Reads  (view any size file with low memory overhead)
*      Relative Position Indicator
*      Panning
*      User-defineable Color Blocks
*      Search and Repeat Search
*      Handles all Video Screen sizes (looks nice in 40*132)
*      100% Clipper
*
*      Compile Clipper 5.x -n
****************************************************************************/
#include 'inkey.ch'
#include 'common.ch'
// For 5.01a users who do not yet have the DEFAULT command:
//#xcommand DEFAULT <v1> TO <x1> [, <vn> TO <xn> ]                        ;
   //          =>                                                            ;
   //          IF <v1> == NIL ; <v1> := <x1> ; END                           ;
   //          [; IF <vn> == NIL ; <vn> := <xn> ; END ]
/***************************************************************************
* This is the default maximum line length
***************************************************************************/
#define MAXBROWSELENGTH 255

/***************************************************************************
* These defines are used by the LineDisp function to define the
* elements of the display color as written by Todd C. MacDonald.
****************************************************************************/
#define CODE_LEN    3
#define COLOR_CODE  1
#define COLOR_SET   2
#define COLOR_DELIM '~~'

/****************************************************************************
* Sample usage
* NOTE - the FileBrowse function has no error trapping built in for
* opening the file to browse - it assumes you have verified the existence
* of a valid file. Although designed to browse a standard text file
* with a line termination string of CR LF, it is also error-trapped
* so it will not choke on a binary file (but don't expect any meaningful
* display :).
****************************************************************************/

*!*****************************************************************************
*!
*!       Function: FBROWSE()
*!
*!      Called by: SNAP.PRG                          
*!               : WELCOME.PRG                       
*!               : FILE_VIEW()        (function  in DBF_OPEN.PRG)
*!
*!          Calls: FILEBROWSE()       (function  in FBROWSE.PRG)
*!
*!*****************************************************************************
function FBROWSE( CFILENAME )
local OLDBLINK

/**************************************************************************
* aDispColors variable:
* ~~7 ~~R This is modeled from Todd C. MacDonald's example provided in his
* ~~7 ~~R LineDisp() function. The actual delimiter is left to you and you
* ~~7 ~~R can use whatever you want just be sure it is guaranteed unique
* ~~7 ~~R and is properly inserted in the COLOR_DELIM define above.
* ~~7 ~~R Todd used the double tilde hence I am simply carrying forward
* ~~7 ~~R his convention. The double tilde in this comment block will
* ~~7 ~~R cause these lines to be displayed in BRIGHT WHITE on RED if
* ~~7 ~~R you compile, link and browse this program file. The double
* ~~7 ~~R tilde character will be stripped before display.
* ~~7 ~~N This line will be BLACK on RED.
* ~~7              (And the whole thing is in a semi-shadow)
****************************************************************************/
local ADISPCOLORS := { { '~~R', 'W+/R'  } , { '~~7', 'W/N'   } , ;
{ '~~N', 'N/R'   } , { '~~D', NIL     } }

/***************************************************************************
* bApplyText:
* This code block will be applied to all lines of text before they are
* stored in the display array as follows: cLine = Eval( bApplyText , cLine )
* A standard usage of this block would be to strip imbedded printer control
* codes for display of the file.
* Credited to Todd C. MacDonald, who encloses the printer codes in
* delimiters, strips the entire code for display, then strips just the
* delimiters to allow the user to immediately send the same file to the
* printer with printer control codes intact.
****************************************************************************/
local BAPPLYTEXT := { |X| Strtran( x ,"Function" , "FUNCTION" ) }

set scoreboard Off
if CFILENAME == NIL
   ? 'Browse <cFileName>'
elseif !file( CFILENAME )
   ? CFILENAME + ' not found!'
else
   OLDBLINK = setblink(.f.)
   FILEBROWSE(CFILENAME,0,0,Maxrow(),maxcol(),'w+/b','b+/b','b+/w','r+/bg',;
      BAPPLYTEXT,ADISPCOLORS)
   setblink( OLDBLINK )
endif
return NIL

* End of commercial announcement

/****************************************************************************
* Syntax          FileBrowse( <cFileName> , [<nTop>] , [<nLeft>] ,
*                     [<nBottom>] , [<nRight>] , [<cWinColor>] ,
*                     [<cBoxColor>] , [<cBarColor>] , [<cButtonColor>] ,
*                     [<bApplyText>] , [<aDispColors>] )
*
* Arguments       <cFileName> DOS file to browse. Provide your own
*                     error trapping prior to calling FileBrowse().
*                     This function assumes a valid, readable file.
*                 <nTop>, <nLeft>, <nBottom> , <nRight> browse window
*                     coordinates; If not specified, defaults to 0,0,
*                     MaxRow(),MaxCol()
*                 <cWinColor> , <cBoxColor> , <cBarColor>, <cButtonColor>
*                     Clipper color strings; If not specified, defaults to
*                     SetColor() , Setcolor() , "W/N" , "N/W" respectively.
*                 <bApplyText>  An optional code block which will be
*                     applied to each line of text as it is extracted
*                     from the file allowing control codes, etc. to be
*                     removed prior to display
*                 <aDispColors> An optional array of codes and Clipper color
*                     strings to affect the display colors. Each element of
*                     the array contains { cCode , cColorString }. Each
*                     cCode used must begin with a common delimiter and
*                     contain a unique identifier. Example:
*                      { { '~~WR' , 'W+/R' } , { ~~BW , 'B/W' } } where
*                     ~~ is the common delimiter and WR or BW is the
*                     unique identifier portion of each cCode.
*                     Imbed the cCode in the text file to display the text
*                     in the corresponding color. Each color change is
*                     only applicable to the end of the  current line or
*                     another cCode is encountered; subsequent lines revert
*                     to the default cWinColor. If not specified, all
*                     output will be in the cWinColor.
* Returns         NIL
*
* Description     Pure Clipper text file browser
*
* Calls           NONE
*
* Notes           Beats the pants off of the similar tBrowse implementation
****************************************************************************/
*!*****************************************************************************
*!
*!       Function: FILEBROWSE()
*!
*!      Called by: FBROWSE()          (function  in FBROWSE.PRG)
*!
*!          Calls: DOWNFILLARRAY      (procedure in FBROWSE.PRG)
*!               : SCROLLBAR          (procedure in FBROWSE.PRG)
*!               : LINEDISP           (procedure in FBROWSE.PRG)
*!               : HIGHLIGHT          (procedure in FBROWSE.PRG)
*!               : SKIPDOWN           (procedure in FBROWSE.PRG)
*!               : SKIPUP             (procedure in FBROWSE.PRG)
*!               : UPFILLARRAY        (procedure in FBROWSE.PRG)
*!               : SEARCH             (procedure in FBROWSE.PRG)
*!               : CHECKDISP          (procedure in FBROWSE.PRG)
*!               : SHOWHELP()         (function  in FBROWSE.PRG)
*!
*!*****************************************************************************
function FILEBROWSE(CFILENAME,NTOP,NLEFT,NBOTTOM,NRIGHT,CWINCOLOR,;
   CBOXCOLOR,CBARCOLOR,CBUTTONCOLOR,BAPPLYTEXT,ADISPCOLORS)
// the browse variables
local ALINES
local AWINBUFF := {Savescreen(),setcolor(CWINCOLOR),row(),col(),setcursor(0)}
local LAPPLYBLOCK := (BAPPLYTEXT # NIL .and. valtype(BAPPLYTEXT) == "B")
local LHITBOTTOM := .f.
local NCURRENTLEFT := 1
local NHANDLE := fopen( CFILENAME , 0 )
local NKEY := 0
local NLASTOFFSET := 0
local NLASTLINE := 0
local NLEFTBROWSE := 1
local NLENGTHBROWSE
local NMAXRIGHT := 0
local NSIZE := Fseek( NHANDLE , 0 , 2 )
// for the ScrollBar
local NROW
local NBARTOP
local NBARBOTTOM
// for the search function
local CLOOKTEXT := ''
local NHIGHLITELINE := 0
local NHIGHLITEOFFSET := 0
local NLASTFIND := 0
local NLOOKLEN := 0
local NTOTALFOUND := 0
// move back to the top of the file
Fseek(NHANDLE,0)
// set the defaults
default NTOP to 0
default NLEFT to 0
default NBOTTOM to Maxrow()
default NRIGHT to maxcol()
default CWINCOLOR to setcolor()
default CBOXCOLOR to setcolor()
default CBARCOLOR to 'W/N'
default CBUTTONCOLOR to  'N/W'
// Set other vars; NOTE: if you change the window look at these carefully
NROW = NTOP+1
NBARTOP = NTOP+1
NBARBOTTOM = NBOTTOM-1
NLASTLINE = NBOTTOM - NTOP - 1
NLENGTHBROWSE = (NRIGHT - NLEFT - 2 )
// declare and fill the array
ALINES = array( NLASTLINE )
Afill( ALINES , {'',0,0} )
DOWNFILLARRAY(NHANDLE,ALINES,1,NLASTLINE,LAPPLYBLOCK,BAPPLYTEXT)
// paint the screen
dispbegin()
Scroll(NTOP,NLEFT,NBOTTOM,NRIGHT)
dispbox(NTOP,NLEFT,NBOTTOM,NRIGHT,,CBOXCOLOR)
SCROLLBAR(.t.,ALINES[1,2],NLASTOFFSET,NSIZE,NLASTLINE,;
   NBARTOP,NRIGHT,NBARBOTTOM,@NROW,CBARCOLOR,CBUTTONCOLOR)
dispend()
while .t.
   // reset default values
   NMAXRIGHT := NLASTOFFSET := 0
   // display the screen
   dispbegin()
   LINEDISP(ALINES,NTOP,NLEFT,NLEFTBROWSE,NLENGTHBROWSE,;
      @NMAXRIGHT,@NLASTOFFSET,ADISPCOLORS,NLASTLINE,CWINCOLOR)
   if NHIGHLITELINE # 0  // highlight the search text
      NHIGHLITELINE = HIGHLIGHT(CLOOKTEXT,NHIGHLITEOFFSET-NLEFTBROWSE+1,;
         NTOP+NHIGHLITELINE,NLEFT+1)
   endif
   SCROLLBAR(.f.,ALINES[1,2],NLASTOFFSET,NSIZE,NLASTLINE,NBARTOP,NRIGHT,;
      NBARBOTTOM,@NROW,CBARCOLOR,CBUTTONCOLOR) // update the scroll bar
   dispend()
   clear typeahead // I like this thing to stop when I stop pressing a key
   do case
   case (NKEY := inkey(0)) == K_ESC
      exit
   case NKEY == K_DOWN
      SKIPDOWN(NHANDLE,ALINES,NLASTLINE,1,NSIZE,LAPPLYBLOCK,BAPPLYTEXT)
   case NKEY == K_UP
      SKIPUP(NHANDLE,ALINES,LAPPLYBLOCK,BAPPLYTEXT)
   case NKEY == K_PGUP
      if !UPFILLARRAY(NHANDLE,ALINES,NLASTLINE,NSIZE,LAPPLYBLOCK,BAPPLYTEXT)
         Fseek( NHANDLE , 0 )
         Afill( ALINES , {'',0,0} )
         DOWNFILLARRAY(NHANDLE,ALINES,1,NLASTLINE,LAPPLYBLOCK,BAPPLYTEXT)
      endif
   case NKEY == K_PGDN
      SKIPDOWN(NHANDLE,ALINES,NLASTLINE,NLASTLINE,NSIZE,;
         LAPPLYBLOCK,BAPPLYTEXT)
   case NKEY == K_LEFT .and. NLEFTBROWSE > 1
      NLEFTBROWSE --
   case NKEY == K_RIGHT .and. NLEFTBROWSE < ( NMAXRIGHT - NLENGTHBROWSE )
      NLEFTBROWSE ++
   case NKEY == K_HOME .and. NLEFTBROWSE > 1
      NLEFTBROWSE = 1
   case NKEY == K_END .and. NLEFTBROWSE < ( NMAXRIGHT - NLENGTHBROWSE )
      NLEFTBROWSE = ( NMAXRIGHT - NLENGTHBROWSE )
   case NKEY == K_CTRL_PGUP //.and. aLines[1,2] # 0
      Fseek( NHANDLE , 0 )
      DOWNFILLARRAY(NHANDLE,ALINES,1,NLASTLINE,LAPPLYBLOCK,BAPPLYTEXT)
   case NKEY == K_CTRL_PGDN
      ALINES[1,2] = Fseek( NHANDLE , 0 , 2 )
      if !UPFILLARRAY(NHANDLE,ALINES,NLASTLINE,NSIZE,LAPPLYBLOCK,BAPPLYTEXT)
         Fseek( NHANDLE , 0 )
         Afill( ALINES , {'',0,0} )
         DOWNFILLARRAY(NHANDLE,ALINES,1,NLASTLINE,LAPPLYBLOCK,BAPPLYTEXT)
      endif
   case NKEY == K_TAB .and. NLEFTBROWSE < ( NMAXRIGHT - NLENGTHBROWSE )
      NLEFTBROWSE = min(NLEFTBROWSE+NLENGTHBROWSE,NLEFTBROWSE+5)
   case NKEY == K_SH_TAB .and. NLEFTBROWSE > 1
      NLEFTBROWSE = max(0,NLEFTBROWSE-5)
   case NKEY == K_ALT_F .or. NKEY == K_ALT_R
      if (NLASTOFFSET := SEARCH(NHANDLE,if(NKEY==K_ALT_F,1,2),;
            @NLASTFIND,@CLOOKTEXT,@NLOOKLEN,@NHIGHLITEOFFSET,;
            NBOTTOM,@NTOTALFOUND) ) > 0
         DOWNFILLARRAY(NHANDLE,ALINES,1,NLASTLINE,@LHITBOTTOM,;
            LAPPLYBLOCK,BAPPLYTEXT)
         if LHITBOTTOM .or. Ascan(ALINES,{|X|!empty(x[1])}) = 0
            // we hit bottom or no text in any line
            Afill( ALINES , {'',0,0} )
            ALINES[1,2] = Fseek( NHANDLE , 0 , 2 )
            if !UPFILLARRAY(NHANDLE,ALINES,NLASTLINE,NSIZE,;
                  LAPPLYBLOCK,BAPPLYTEXT)
               Fseek( NHANDLE , 0 )
               Afill( ALINES , {'',0,0} )
               // this is not really as redundant as it may first appear
               DOWNFILLARRAY(NHANDLE,ALINES,1,NLASTLINE,;
                  LAPPLYBLOCK,BAPPLYTEXT)
            endif
         endif
         NHIGHLITELINE = CHECKDISP(CLOOKTEXT,ALINES,NLASTOFFSET,;
            NHIGHLITEOFFSET,@NLEFTBROWSE,NLENGTHBROWSE)
      endif
   case NKEY == K_ALT_H .or. NKEY == K_F1
      SHOWHELP()
   case setkey( NKEY ) # NIL
      eval( setkey( NKEY) )
   endcase
end
Fclose( NHANDLE )
setcolor( AWINBUFF[2]) // next 4 lines are my screen restore stuff
Restscreen(0,0,Maxrow(),maxcol(),AWINBUFF[1])
setpos(AWINBUFF[3],AWINBUFF[4])
setcursor(AWINBUFF[5])
return NIL

/****************************************************************************
* Displays the lines of text contained in the array
****************************************************************************/

*!*****************************************************************************
*!
*!      Procedure: LINEDISP
*!
*!      Called by: FILEBROWSE()       (function  in FBROWSE.PRG)
*!
*!*****************************************************************************
static function LINEDISP(ALINES,NTOP,NLEFT,NLEFTBROWSE,NLENGTHBROWSE,;
   NMAXRIGHT,NLASTOFFSET,ADISPCOLORS,NLASTLINE,;
   CWINCOLOR)
// The LOCAL vars for Todd's part of this function
local CCOLORCODE
local CCOLORSET
local CLINE
local COUTPUT
local NCOLORCODE
local NCODEPOS
local NDIFF
local NKEY
local NLENGTH
local NLINE
local NLINEPOS

NLEFT ++ // this needs to be incremented for both display methods
if ADISPCOLORS = NIL // no colors defined so use the standard display method
   aeval(ALINES,{ |X,Y|DEVPOS(NTOP+Y,NLEFT),;
      devout(pad(substr(x[1],NLEFTBROWSE),NLENGTHBROWSE)),;
      NMAXRIGHT   := max(NMAXRIGHT,len(x[1])),;
      NLASTOFFSET := max(NLASTOFFSET,x[3])})
else
/*************************************************************************
* This is an original work by Todd C. MacDonald and is hereby
* placed in the public domain.
*
* The framework for this portion of the LineDisp function was
* graciously provided by Todd C. MacDonald. It is included here with
* modifications to use the established variables and conventions in
* FileBrowse - Tom Claffy 9-28-93
**************************************************************************/
   
   NTOP ++
   for NLINE = 1 to NLASTLINE
      setpos( NTOP++, NLEFT )
      CLINE = ALINES[ nLine , 1 ]
      NLINEPOS = 1
      CCOLORSET = CWINCOLOR
      NLENGTH = len( CLINE )
      while (COLOR_DELIM $ CLINE)
         NCODEPOS = at( COLOR_DELIM , CLINE )
         COUTPUT = left( CLINE , NCODEPOS - 1 )
         // strip the beginning of the line if we are panned right
         // this must be done as a color code may be in this part
         // of the line so we must process the characters but cannot
         // display any characters until we reach the first virtual column
         // of the display window
         if NLINEPOS < NLEFTBROWSE
            NDIFF = min( len( COUTPUT ) , NLEFTBROWSE - NLINEPOS )
            NLINEPOS += NDIFF
            COUTPUT = substr( COUTPUT , NDIFF + 1 )
         endif
         // output line up to code position in current color
         devout( COUTPUT , CCOLORSET )
         // strip off text just displayed
         CLINE =  substr( CLINE, NCODEPOS )
         // Set new color based on color code.  If the color code
         // is not found in the array, the code is not stripped out
         // (therefore the code itself gets displayed in the output
         // text).  If the code is found but the color value is nil,
         // the color is set to the default.
         CCOLORCODE = left( CLINE, CODE_LEN )
         if ( NCOLORCODE := Ascan( ADISPCOLORS, ;
               { | a | a[ COLOR_CODE ] == CCOLORCODE } ) ) # 0
            if ( CCOLORSET := ADISPCOLORS[ nColorCode, COLOR_SET ] ) =  NIL
               CCOLORSET = CWINCOLOR
            endif
            // strip off color code
            CLINE = substr( CLINE, CODE_LEN + 1 )
            // keep track of the line length
            NLENGTH -= (CODE_LEN +1)
         else
            // strip off color code prefix
            CLINE = substr( CLINE, CODE_LEN )
            // keep track of the line length
            NLENGTH -= CODE_LEN
         endif
      end
      // cut to the left column if we are not already there
      if NLINEPOS < NLEFTBROWSE
         CLINE = substr( CLINE , NLEFTBROWSE - NLINEPOS + 1)
      endif
      // output remainder of line
      devout( pad( CLINE , NLEFT + NLENGTHBROWSE - col() ) , CCOLORSET )
      // set the system counters
      NMAXRIGHT   = max(NMAXRIGHT,NLENGTH)
      NLASTOFFSET = max(NLASTOFFSET,ALINES[nLine,3])
   next
endif
return NIL

/***************************************************************************
* Justify the display before displaying the found text
* The find column may be out of view left or right
****************************************************************************/
*!*****************************************************************************
*!
*!      Procedure: CHECKDISP
*!
*!      Called by: FILEBROWSE()       (function  in FBROWSE.PRG)
*!
*!*****************************************************************************
static function CHECKDISP(CLOOKTEXT,ALINES,NLASTOFFSET,;
   NHIGHLITEOFFSET,;
   NLEFTBROWSE,NLENGTHBROWSE)
local NLENGTH := len(alltrim(CLOOKTEXT))
if NHIGHLITEOFFSET < NLEFTBROWSE  // the find text is out of scope left
   NLEFTBROWSE = NHIGHLITEOFFSET
elseif NLEFTBROWSE+NHIGHLITEOFFSET+NLENGTH > ;
      NLEFTBROWSE+NLENGTHBROWSE  // the find text is out of scope right
   NLEFTBROWSE = (NHIGHLITEOFFSET+NLENGTH-NLENGTHBROWSE )
endif
return (Ascan(ALINES,{ |X| x[2] <= NLASTOFFSET .and. x[3] > NLASTOFFSET}))

/***************************************************************************
* Highlight the found text
****************************************************************************/
*!*****************************************************************************
*!
*!      Procedure: HIGHLIGHT
*!
*!      Called by: FILEBROWSE()       (function  in FBROWSE.PRG)
*!
*!*****************************************************************************
static function HIGHLIGHT(CLOOKTEXT,NOFFSET,NROW,NCOL)
local NLENGTH := len(alltrim(CLOOKTEXT))
local CSCREEN := Savescreen(NROW,NCOL+NOFFSET-1,;
   NROW,NCOL+NOFFSET+NLENGTH-2)
// use this variable to set the highlight color you want to use
local CCHAR := if( chr(7) $ CSCREEN,chr(112),chr(7))
Restscreen(NROW,NCOL+NOFFSET-1,NROW,NCOL+NOFFSET+NLENGTH-2,;
   transform(CSCREEN,replicate(("X"+CCHAR),NLENGTH)))
return 0

/***************************************************************************
* Clean up the line by removing control characters
* and apply the user-defined block if applicable
* cLine is passed to this function by reference
* thus no return value
****************************************************************************/
*!*****************************************************************************
*!
*!      Procedure: LINECLEAN
*!
*!      Called by: DOWNFILLARRAY      (procedure in FBROWSE.PRG)
*!               : UPFILLARRAY        (procedure in FBROWSE.PRG)
*!               : SKIPUP             (procedure in FBROWSE.PRG)
*!
*!*****************************************************************************
static function LINECLEAN( CLINE , LAPPLYBLOCK , BAPPLYTEXT )
CLINE = Strtran( CLINE , chr(10) , ' ' )   // LF
CLINE = Strtran( CLINE , chr(12) , ' ' )   // FF
CLINE = Strtran( CLINE , chr(13) , ' ' )   // CR
CLINE = Strtran( CLINE , chr(26) , ' ' )   // EOF
if LAPPLYBLOCK
   CLINE = eval( BAPPLYTEXT , CLINE )
endif
return NIL

/****************************************************************************
* Search for line-feed or form-feed mark - return the first offset
* The extraordinary error trapping is included here to prevent
* lock-up when a binary file is browsed, i.e. a loop can be
* caused by the eof mark not being found
****************************************************************************/
*!*****************************************************************************
*!
*!      Procedure: FIRSTEOLMARK
*!
*!      Called by: DOWNFILLARRAY      (procedure in FBROWSE.PRG)
*!               : UPFILLARRAY        (procedure in FBROWSE.PRG)
*!
*!*****************************************************************************
static function FIRSTEOLMARK(CBUFFER)
local NFFMARK := at(chr(12),CBUFFER)
local NLFMARK := at(chr(10),CBUFFER)
local NRETURN := 0
if NFFMARK > 0 .and. NLFMARK > 0
   NRETURN := min( min(NFFMARK,NLFMARK) , MAXBROWSELENGTH )
elseif NFFMARK > 0 .or. NLFMARK > 0
   NRETURN := min( max(NFFMARK,NLFMARK) , MAXBROWSELENGTH )
elseif len( CBUFFER ) >= MAXBROWSELENGTH
   NRETURN := MAXBROWSELENGTH
endif
return NRETURN

/***************************************************************************
* fill the array traversing down the file
****************************************************************************/
*!*****************************************************************************
*!
*!      Procedure: DOWNFILLARRAY
*!
*!      Called by: FILEBROWSE()       (function  in FBROWSE.PRG)
*!               : SKIPDOWN           (procedure in FBROWSE.PRG)
*!
*!          Calls: FIRSTEOLMARK       (procedure in FBROWSE.PRG)
*!               : LINECLEAN          (procedure in FBROWSE.PRG)
*!
*!*****************************************************************************
static function DOWNFILLARRAY(NHANDLE,ALINES,NSTART,NEND,;
   LAPPLYBLOCK,BAPPLYTEXT,LHITBOTTOM)
local NBLOCK := 1024
local CBUFFER
local CLINE
local NEOL
local NCOUNTER := NSTART - 1
local NFILEPOS := Fseek( NHANDLE , 0 , 1 )
local NLINELENGTH
local NLOOPCOUNTER := 0
local NBYTESREAD
LHITBOTTOM = .f.
// while the array is not full
while NCOUNTER < NEND
   // reposition to current file position
   Fseek( NHANDLE , NFILEPOS , 0 )
   // assign a buffer and read the file
   CBUFFER = space( NBLOCK )
   if (NBYTESREAD := Fread( NHANDLE , @CBUFFER , NBLOCK )) # NBLOCK
      NLOOPCOUNTER ++
   endif
   // check for eof
   if NBYTESREAD < 1 .or. NLOOPCOUNTER > 2
      LHITBOTTOM = .t.
      exit
   else
      while NCOUNTER < NEND .and. (NEOL := FIRSTEOLMARK( CBUFFER )) # 0
         NCOUNTER ++
         // extract the line
         CLINE = left( CBUFFER , NEOL )
         // strip the line from the buffer
         CBUFFER = substr(CBUFFER,NEOL+1)
         // save the length
         NLINELENGTH = len( CLINE )
         // clean it up
         LINECLEAN( @CLINE , LAPPLYBLOCK , BAPPLYTEXT )
         // store it in the array
         ALINES[nCounter] = { CLINE , NFILEPOS , NFILEPOS + NLINELENGTH }
         // keep the pointer current
         NFILEPOS += NLINELENGTH
      end
      // test for eof
      LHITBOTTOM = (NCOUNTER < NEND)
   endif
end
return NIL

/***************************************************************************
* add one line to the bottom the array
****************************************************************************/
*!*****************************************************************************
*!
*!      Procedure: SKIPDOWN
*!
*!      Called by: FILEBROWSE()       (function  in FBROWSE.PRG)
*!
*!          Calls: DOWNFILLARRAY      (procedure in FBROWSE.PRG)
*!               : UPFILLARRAY        (procedure in FBROWSE.PRG)
*!
*!*****************************************************************************
static function SKIPDOWN(NHANDLE,ALINES,NLASTLINE,NNUMLINES,NSIZE,;
   LAPPLYBLOCK,BAPPLYTEXT)
local LHITBOTTOM := .f.
if ALINES[ nLastLine , 3 ] # 0
   // position file to last line offset
   Fseek( NHANDLE , ALINES[ nLastLine , 3 ] , 0 )
   adel( ALINES , 1 )
   ALINES[nLastLine] = {'',0,0}
   // get the next line
   DOWNFILLARRAY(NHANDLE,ALINES,NLASTLINE+1-NNUMLINES,NLASTLINE,;
      LAPPLYBLOCK,BAPPLYTEXT,@LHITBOTTOM)
   if LHITBOTTOM .or. Ascan(ALINES,{|X|!empty(x[1])}) = 0
      // we hit bottom or no text in any line
      ALINES[1,2] = Fseek( NHANDLE , 0 , 2 )
      UPFILLARRAY(NHANDLE,ALINES,NLASTLINE,NSIZE,LAPPLYBLOCK,BAPPLYTEXT)
   endif
endif
return NIL

/***************************************************************************
* fill the array traversing up ... the tricky one ...
* This function was originally coded with the RAT function
* Tests show the AT func to be about 10 times faster than RAT
* thus the overhead of the aTemp with the AT function as an offset map
****************************************************************************/
*!*****************************************************************************
*!
*!      Procedure: UPFILLARRAY
*!
*!      Called by: FILEBROWSE()       (function  in FBROWSE.PRG)
*!               : SKIPDOWN           (procedure in FBROWSE.PRG)
*!
*!          Calls: FIRSTEOLMARK       (procedure in FBROWSE.PRG)
*!               : LINECLEAN          (procedure in FBROWSE.PRG)
*!
*!*****************************************************************************
static func UPFILLARRAY(NHANDLE,ALINES,NEND,NSIZE,LAPPLYBLOCK,BAPPLYTEXT)
local AOFFSETS
local CBUFFER
local CFIRSTCHAR
local CLINE := ''
local NBLOCK := 1024
local NCOUNTER := 0
local NEOL
local NFILEPOS := Fseek( NHANDLE , 0 , 1 )
local NTEMPCOUNT
local LEOF :=  (NSIZE = NFILEPOS )
local LRETURN := .t.
while NCOUNTER < NEND
   // assign some defaults and read the file
   NBLOCK = min(NBLOCK,ALINES[1,2])
   NFILEPOS = ALINES[1,2] - NBLOCK
   Fseek( NHANDLE , NFILEPOS , 0 )
   CBUFFER = space( NBLOCK )
   if Fread( NHANDLE , @CBUFFER , NBLOCK ) = 0
      exit
   else
      if NFILEPOS = 0
         AOFFSETS = {1}
      else
         AOFFSETS = { }     // discard the first line - it is a fragment
      endif
      // map the lines into a temp array
      NTEMPCOUNTER = 0
      while (NEOL := FIRSTEOLMARK( substr( CBUFFER , NTEMPCOUNTER + 1 ))) # 0
         NTEMPCOUNTER += NEOL
         Aadd( AOFFSETS , NTEMPCOUNTER )
      end
      // pick up the last line if eof and it does not end w/ LF CR etc.
      if LEOF .and. NTEMPCOUNTER < NBLOCK
         Aadd( AOFFSETS , NBLOCK )
      endif
      NTEMPCOUNTER = len( AOFFSETS )
      // fill the array
      while NCOUNTER < NEND .and. NTEMPCOUNTER > 1
         NCOUNTER ++
         NTEMPCOUNTER --
         CLINE = substr(CBUFFER,AOFFSETS[nTempCounter]+1)
         // strip the first char if it is a control char
         // going down we don't have this problem as it is at the end and
         // we don't care
         CFIRSTCHAR = left( CLINE , 1 )
         if CFIRSTCHAR = chr(10) .or. CFIRSTCHAR = chr(12) .or. ;
               CFIRSTCHAR = chr(13) .or. CFIRSTCHAR = chr(26)
            CLINE = substr( CLINE , 2 )
         endif
         // clean it up
         LINECLEAN( @CLINE , LAPPLYBLOCK , BAPPLYTEXT )
         // store the line and it's parameters
         Ains( ALINES , 1 )
         ALINES[1] = {CLINE,;
            NFILEPOS+AOFFSETS[nTempCounter],;
            NFILEPOS+AOFFSETS[nTempCounter+1]}
         // strip the line from the buffer
         CBUFFER = left(CBUFFER,AOFFSETS[nTempCounter])
      end
      // if nCounter < nEnd we ran out of lines
      // return .f. and fill the array from the top
      LRETURN = (NCOUNTER = NEND )
      if NFILEPOS = 0
         if ALINES[1,2] = 1 // assign BOF status explicitly for the other
            ALINES[1,2] = 0 // functions
         endif
         exit  // we're done here
      endif
   endif
end
return LRETURN

/***************************************************************************
* add one line to the top of the array
****************************************************************************/
*!*****************************************************************************
*!
*!      Procedure: SKIPUP
*!
*!      Called by: FILEBROWSE()       (function  in FBROWSE.PRG)
*!
*!          Calls: LINECLEAN          (procedure in FBROWSE.PRG)
*!
*!*****************************************************************************
static func SKIPUP(NHANDLE,ALINES,LAPPLYBLOCK,BAPPLYTEXT)
local NBLOCK
local CBUFFER
local CLINE := ''
local LBOF := .f.
local NEOL
local NFILEPOS
local NLENGTH
if ALINES[1,2] #0
   NBLOCK = min(512,ALINES[1,2])
   NFILEPOS = ALINES[1,2] - NBLOCK
   LBOF = (NFILEPOS <= 1 )
   Fseek( NHANDLE , NFILEPOS , 0 )
   CBUFFER = space( NBLOCK )
   if !Fread( NHANDLE , @CBUFFER , NBLOCK ) = 0
      // get past the first eol mark
      NEOL = max( Rat(chr(12),CBUFFER), Rat(chr(10),CBUFFER) )
      if NEOL # 0
         CLINE = right( CBUFFER , len(CBUFFER)-NEOL+1 )
         CBUFFER = left(CBUFFER,NEOL-1)
      endif
      // get the last line in the buffer
      NEOL = max( Rat(chr(12),CBUFFER), Rat(chr(10),CBUFFER) )
      if NEOL == 0 .and. !LBOF
         NEOL = min(NBLOCK,MAXBROWSELENGTH)
      elseif len(CBUFFER) - NEOL > MAXBROWSELENGTH
         NEOL = MAXBROWSELENGTH
      endif
      // update the file position
      NFILEPOS += NEOL
      // get the line
      CLINE = ( right(CBUFFER,len(CBUFFER)-NEOL) + CLINE )
      // save the real length of the line in the file
      NLENGTH = len( CLINE )
      // clean it up
      LINECLEAN( @CLINE , LAPPLYBLOCK , BAPPLYTEXT )
      // store the line and it's parameters
      Ains( ALINES , 1 )
      ALINES[1] = {CLINE,NFILEPOS,NFILEPOS+NLENGTH}
   endif
endif
return NIL

/***************************************************************************
* Display a status bar based on the relative position in the file
* Not 100 % accurate but a fair representation to pacify the mere mortals
****************************************************************************/
*!*****************************************************************************
*!
*!      Procedure: SCROLLBAR
*!
*!      Called by: FILEBROWSE()       (function  in FBROWSE.PRG)
*!
*!*****************************************************************************
static function SCROLLBAR(LPAINT,NSTART,NEND,NSIZE,NLASTLINE,NBARTOP,NRIGHT,;
   NBARBOTTOM,NROW,CBARCOLOR,CBUTTONCOLOR)
local NCOUNTER
local NMIDDLE
local NPERCENT
if LPAINT  // this only happens on the very first call
   for NCOUNTER = NBARTOP to NBARBOTTOM
      devpos(NCOUNTER,NRIGHT); devout(chr(176),CBARCOLOR)
   next
else // update the button
   // erase the old button - where's the whiteout
   devpos(NROW,NRIGHT); devout(chr(176),CBARCOLOR)
   if NSTART = 0             // figger out the new row
      NROW = NBARTOP         // nRow is passed by ref so it will be updated
   elseif NEND >= NSIZE -1
      NROW = NBARBOTTOM
   else
      NMIDDLE = NSTART + ((NEND - NSTART) / 2 )
      NPERCENT = (NMIDDLE/NSIZE)
      NROW = NBARTOP + int( (NBARBOTTOM - NBARTOP) * NPERCENT )
   endif
   // display the button
   devpos(NROW,NRIGHT); devout( chr(254),CBUTTONCOLOR)
endif
return NIL

/***************************************************************************
* Search Routine By: Phil Barnett  April 3, 1993
* Search and repeat last find
****************************************************************************/
*!*****************************************************************************
*!
*!      Procedure: SEARCH
*!
*!      Called by: FILEBROWSE()       (function  in FBROWSE.PRG)
*!
*!*****************************************************************************
static function SEARCH(NHANDLE,NMODE,NLASTFIND,CLOOKTEXT,NLOOKLEN,;
   NHIGHLITEOFFSET,NBOTTOM,NTOTALFOUND)
local CBUFFER
local CSAVEIT := Savescreen(NBOTTOM,1,NBOTTOM,60)
local NBYTESREAD
local NBLOCK := 4096  // Bigger is probably faster up to ~10000 bytes
local NTHISOFFSET
local NOFFSET := 0
local NLOOP := 0
local GETLIST := {}
local CMINIBUFF
local NMBSIZE
local NEOLAT
local CHOLDBACK := ''
if NMODE == 1     // initiate search (ALT_F)ind
   NTOTALFOUND = 0
   NLASTFIND = -1
   CLOOKTEXT = pad(CLOOKTEXT,25)
   @ NBOTTOM,1 say space(49)
   @ NBOTTOM,2 say 'Enter Search Phrase' get CLOOKTEXT picture '@K@!'
   setcursor( if(Readinsert(),2,1 ))
   read
   setcursor(0)
   Restscreen(NBOTTOM,1,NBOTTOM,60,CSAVEIT)
   if lastkey() == 27 .or. empty(CLOOKTEXT)
      return 0
   endif
   CLOOKTEXT = upper(alltrim(CLOOKTEXT))
   NLOOKLEN = len(CLOOKTEXT)
endif
Fseek(NHANDLE,NLASTFIND+1,0)  // Position filepointer to starting location
NBYTESREAD = NBLOCK          // setup loop entry
do while NBYTESREAD == NBLOCK
   // Get and prepare the Character Buffer
   CBUFFER = space(NBLOCK)
   NBYTESREAD = Fread(NHANDLE,@CBUFFER,NBLOCK)
   if NBYTESREAD < NBLOCK
      CBUFFER = left(CBUFFER,NBYTESREAD)
   endif
   CBUFFER = CHOLDBACK + upper(CBUFFER)
   NTHISOFFSET = at(CLOOKTEXT,CBUFFER)
   if !empty(NTHISOFFSET)
      // position the File Pointer to the find.
      NLASTFIND += ( NLOOP * NBLOCK ) + NTHISOFFSET - len( CHOLDBACK )
      Fseek( NHANDLE, NLASTFIND , 0)
      // Now, locate the beginning of the line
      // ( It might not be in current buffer, so make a new minibuffer )
      NMBSIZE = min( NLASTFIND, 512 )
      CMINIBUFF = space(NMBSIZE)
      Fseek( NHANDLE,-NMBSIZE,1)
      Fread(NHANDLE,@CMINIBUFF,NMBSIZE)
      NEOLAT = max( Rat(chr(12),CMINIBUFF), Rat(chr(10),CMINIBUFF) )
      if NEOLAT > 0
         Fseek( NHANDLE, NEOLAT-NMBSIZE,1)
      endif
      NOFFSET = Fseek(NHANDLE,0,1)
      NHIGHLITEOFFSET = (NLASTFIND - NOFFSET + 1)
      NTOTALFOUND ++
      // We found one so...
      exit
   endif
   CHOLDBACK = right(CBUFFER,NLOOKLEN - 1)
   NLOOP ++
enddo
if NOFFSET == 0
   @ NBOTTOM,1 say ' ' +ltrim(trim(str(NTOTALFOUND)))+;
      ' Occurences Found - No More Finds (Press Any Key) '
   Tone(100,1)
   inkey(0)
   Restscreen(NBOTTOM,1,NBOTTOM,60,CSAVEIT)
endif
return NOFFSET

/***************************************************************************
* Help Routine By: Bryan Smith November 1994
****************************************************************************/
*!*****************************************************************************
*!
*!       Function: SHOWHELP()
*!
*!      Called by: FILEBROWSE()       (function  in FBROWSE.PRG)
*!
*!*****************************************************************************
function SHOWHELP()
local TLR:= 1, TLC:= 7, BRR:= 22, BRC:= 71  // top left, bot rt row & col
local CHDR := "INSTRUCTIONS FOR USING THE FILE BROWSER"
local CBAR := replicate(chr(205),3)
local LH := len(CHDR)
local SC := TLC + 7  // start column
local XLM := SC + 3  // extra left margin
local AR := TLR      // active row
local INSCREEN := Savescreen(0,0,Maxrow(),maxcol())
local INCURSOR := set( _SET_CURSOR, 0 )
local INCOLOR := setcolor('w+/g, w+/g,,,w+/g')

Scroll(TLR,TLC,BRR,BRC)
dispbox(TLR,TLC,BRR,BRC,2,'g+/g')
AR+=2 ; @AR,XLM say CHDR
AR+=1 ; @AR,XLM say replicate(chr(205),LH)
AR+=2 ; @AR,SC say "To Exit from Browser, press Escape."
AR+=1 ; @AR,SC say "To move up or down, use up/down cursor keys "
AR+=1 ; @AR,SC say "   or PageUp/PageDown."
AR+=1 ; @AR,SC say "To move to start or end of file, use Ctrl-PageUp"
AR+=1 ; @AR,SC say "   or Ctrl-PageDown."
AR+=1 ; @AR,SC say "To move left or right use left/right cursor keys "
AR+=1 ; @AR,SC say "   or Tab/Shift-Tab or Home/End."
AR+=2 ; @AR,SC say "To Find first time, use Alt-F  (F for Find)."
AR+=1 ; @AR,SC say "To Find same again, use Alt-R  (R for Re-find)."
AR+=1 ; @AR,SC say "   Searches are case-insensitive."
AR+=2 ; @AR,SC say "Note scroll bar on right which shows location in file."
AR+=2 ; @AR,SC say CBAR+ ;
   " PRESS ANY KEY TO RETURN TO THE FILE BROWSER "+CBAR
inkey(0)

Restscreen( 0, 0, Maxrow(), maxcol(), INSCREEN )
set( _SET_CURSOR, INCURSOR )
setcolor( INCOLOR )

return NIL

*: EOF: FBROWSE.PRG
