/*

Directory Wide Structure Comparison (SC.EXE)

Problem: You have clients that have been using your software, and you
have been developing the program. Now it is time to give the client the
new software, but you must insure that the client data files match the
structure that is expected by the new program.

Solution: Create a utility that compares a known set of data structures
with the ones the client is currently using, to inform you which ones
need to be updated, and why.

To use, run SC and enter the drive and path to each directory.

This program will compare the two directories against each other to see
if there have been any structural changes, or if the file is still
needed at all. It will also check to see if there are new files that
need to be added to the client directory.

*/

/*

STRUCOMP.PRG   (abbreviated to SC.PRG by Owen)
Written by Phil Barnett originally

Changes by Owen Parker, oparker@Fox.NSTN.Ca, on Aug 28/95

This documentation follows no rules because I never document things as a
rule. I made the variables very explicitly named but I can barely read
the damn thing, again because I am not so concientious about variable
names by habit.

You can be anywhere when running SC as directories for old and new
copies are entered explicitly by you at run time.

I use a small function to make the old directory path the same as the
new because I always use a subdirectory of the original as my working
place and it lessened the typing time (by 3 seconds - I'm a lazy SOB).
You can change, delete or just type over it if you do not want this auto
default thing to happen.


Finds : Old databases that are no longer needed.
        New databases that need to be added in their entirety.
        Fields whose structures have been changed in some way.
        All missing fields and their specs.
        All extraneous fields.

Option to print to screen, report or file.  Can't overwrite exisiting
file. This should be changed, but I always use screen first then report
if necessary so I didn't bother.

This is a hack program used only by clipper system programmers so no
frills have been added.

This saved me as much time the first couple of times I used it as I
spent re-writing it .  'Nuff said.

It works fine for me in all cases so far but that means nothing.  If you
enhance it please share it by posting it somewhere public (FTP.IAG.NET),
etc.

It is a 'helluva' useful tool.  Thanx Phil.

*/

/*
Addendum by Phil. 02/24/96

I substituted fbrowse for the nanforum library browser to make it so
anyone can compile without any other libs. Plus, it compresses much
smaller this way.

I also included a PBMAKE make script for the rebuild.

Owen did a nice job on the improvements.

I cleaned up the code so it would pass the flags I normally use:

/M /N /W /ES2.

Feel free to change the .MAK to not include compression or use PKLITE or
another file compressor if you want. I have found that the new WWPACK is
better than all the others.

*/

/*

 Modifications by David B. Ferguson  -  Tuesday April 9, 1996

   Cleaned up somewhat visually

   Validates directory entries

   Verifies that data files exist in specified directories

   Fixed double "\" anomaly when root directories entered i.e. A:\ F:\

   Now allows drive only Source and Target entries i.e. C: F: etc.

   Added "/USE" & "/SAVE" Parameters

     /SAVE will save Source and Target directories to SC.DBF
           If SC.DBF does not exist, it will be created.

     /USE  will use saved Source and Target directories if
           SC.DBF exists in the current directory

*/

/*

04/10/96  Phil Barnett

Some more code cleanup

Screen Beautification

Scrolling activity box

Reworded the prompts

*/

#include "inkey.ch"
#include "common.ch"

static OutputType, OutputFile
static WS_ToFile, WS_ToPrnt, WS_ToScr

function sc( Switch )

local getlist := {}
LOCAL cNew_Directory := SPACE(60)
LOCAL cOld_Directory := SPACE(60)
LOCAL Con
local Hld_Screen
local open_screen := SAVESCREEN()
local InRow := ROW()
local InCol := COL()
local SC_Stru := { {"Source","C", 60,0}, {"Target","C", 60,0} }

SETCOLOR( 'GR+/B,N/BG,,,W+/B' )

CLEAR SCREEN
SET SCOREBOARD OFF

DEFAULT Switch to ""

Switch := upper( Switch )

IF "/USE" $ Switch
   IF FILE("SC.DBF")
      USE SC new
      cNew_Directory := sc->Source
      cOld_Directory := sc->Target
      dbclosearea()
   ENDIF
ENDIF

attention( "Structure Compare Utility - Revision 04.96.002", 2 )
attention( "by Phil Barnett with mods by Owen Parker & David B. Ferguson", 23 )

@ 5, 5 SAY "Enter path to newest database files:"
@ 6, 6 GET cNew_Directory PICT "@!" ;
              VALID FILLOLD( cNew_Directory, @cOld_Directory ) ;
               .AND. ! EMPTY( cNew_Directory )
@ 7, 5 SAY "Enter path to compare them to:"
@ 8, 6 GET cOld_Directory PICT "@!K" VALID GoodOldDir( cOld_Directory )

READ

IF LASTKEY() <> K_ESC

   IF "/SAVE" $ Switch

      IF !FILE("SC.DBF")
         DBCREATE( "SC", SC_Stru )
         USE SC new
         DBAPPEND()
         dbclosearea()
      ENDIF

      USE SC new
      sc->Source := cNew_Directory
      sc->Target := cOld_Directory
      dbclosearea()
   ENDIF

   IF LEN( ALLTRIM( cNew_Directory ) ) == 2   // Assumed Drive Only
                                              // so get current directory
      cNew_Directory:=SUBSTR(cNew_Directory, 1, 2) + ;
                      "\"+RTRIM(CURDIR(cNew_Directory))
   ENDIF

   IF RIGHT( ALLTRIM( cNew_Directory ), 1 ) <> "\"
      cNew_Directory := ALLTRIM( cNew_Directory ) + "\"
   ENDIF

   IF LEN( ALLTRIM( cOld_Directory ) ) == 2   // Assumed Drive Only
                                              // so get current directory
      cOld_Directory:=SUBSTR(cOld_Directory, 1, 2) + ;
                      "\"+RTRIM(CURDIR(cOld_Directory))
   ENDIF

   IF RIGHT( ALLTRIM( cOld_Directory ), 1 ) <> "\"
      cOld_Directory := ALLTRIM( cOld_Directory ) + "\"
   ENDIF

   WS_ToScr := WS_ToPrnt := WS_ToFile := .F.

   OutputType := GETOutputType()

   WS_ToScr  := OutputType == "S"
   WS_ToPrnt := OutputType == "P"
   WS_ToFile := OutputType == "F"

   IF WS_ToPrnt
      Con := READYPR()
   ELSE
      Con := ""
   ENDIF

   IF LASTKEY() <> K_ESC .AND. Con <> "Q"
      SAVE SCREEN TO Hld_Screen

      dispbox( 10, 0, 21, 79, "ͻȺ " )

      SET CONSOLE OFF
      SET DEVICE TO PRINT
      SET PRINTER TO ( OutputFile )

      STRUCOMP( cOld_Directory, cNew_Directory )

      SET PRINTER TO
      SET DEVICE TO SCREEN
      SET CONSOLE ON

      RESTORE SCREEN FROM Hld_Screen

      IF WS_ToScr
         FBROWSE( OutputFile )
      ENDIF

      IF WS_ToPrnt
         pop_msg( "Printing...", .f. )
         FSPRINT()
         clr_msg()
      ENDIF

      IF !WS_ToFile
         FERASE( OutputFile )
      ENDIF

   ENDIF
ENDIF

SETPOS( InRow, InCol )
RESTORE SCREEN FROM open_screen
RETURN NIL

/*-------------------------------------------------------------------*/

FUNCTION strucomp( cOld_Directory, cNew_Directory )

LOCAL nField_Structure_Array_Counter, nField_Counter, nFiles_Counter, ;
      nNo_Of_New_Fields, nNo_Of_Old_Fields, aNew_File_Structure_List, ;
      aOld_File_Structure_List, aOld_Files_With_Path, aOld_Just_File_Names, ;
      aNew_Files_With_Path, aNew_Just_File_Names, lPrint_First_Occurrence
local nOld_Fields_Counter, nNew_Fields_Counter
local nOld_Files_Counter, nNew_Files_Counter
local nStructure_Counter
local LGOT_FIRST_OCCURRENCE, cCheck_Field, CCHECK_OLD_FIELD
local LGOT_A_DIFFERENCE, FIRST_IN_FILE, cCheck_File


aOld_Just_File_Names := DIRECTORY( ALLTRIM( cOld_Directory ) + "*.DBF" )
aOld_Files_With_Path := DIRECTORY( ALLTRIM( cOld_Directory ) + "*.DBF" )

FOR nFiles_Counter := 1 TO LEN( aOld_Just_File_Names )
   aOld_Files_With_Path[ nFiles_Counter, 1 ] := ;
     ALLTRIM( cOld_Directory ) + ALLTRIM( aOld_Just_File_Names[ nFiles_Counter, 1 ] )
NEXT

@ PROW() + 1, 2 SAY "Structure Compare Utility "
@ PROW() + 1, 2 SAY "by Phil Barnett with mods by Owen Parker & David B. Ferguson"
@ PROW() + 3, 3 SAY "Newest databases in  -> " + SUBSTR( cNew_Directory, 1, LEN(cNew_Directory)-1 )
@ PROW() + 1, 3 SAY "Compared to files in -> " + SUBSTR( cOld_Directory, 1, LEN(cOld_Directory)-1 )
@ PROW() + 3, 1 SAY ""

FOR nFiles_Counter := 1 TO LEN( aOld_Just_File_Names )

   DBCLOSEALL()

   USE ( aOld_Files_With_Path[ nFiles_Counter, 1 ] ) NEW ALIAS Olddata


   // check to see if the file we are checking is still an active file.

   IF FILE( ALLTRIM( cNew_Directory ) + aOld_Just_File_Names[ nFiles_Counter, 1 ] )
      USE ( ALLTRIM( cNew_Directory ) + aOld_Just_File_Names[ nFiles_Counter, 1 ] ) ;
        NEW ALIAS Newdata
      show( aOld_Files_With_path[ nfiles_counter, 1 ] + ' opened for comparison' )
   ELSE
      @ PROW() + 2, 5 SAY "File " + aOld_Just_File_Names[ nFiles_Counter, 1 ] + ;
          " is no longer needed."
      show( aOld_Files_With_path[ nfiles_counter, 1 ] + ' is no longer needed.' )
      LOOP
   ENDIF

   // then check field count in both tables

   SELECT Newdata
   aNew_File_Structure_List := DBSTRUCT()

   SELECT Olddata
   aOld_File_Structure_List := DBSTRUCT()

   nNo_Of_New_Fields := len( aNew_File_Structure_List )
   nNo_Of_Old_Fields := len( aOld_File_Structure_List )

   show( '' )

   IF nNo_Of_New_Fields <> nNo_Of_Old_Fields

      show( "Field count mismatch  -  New: " + STR( nNo_Of_New_Fields, 3 ) + ;
                                  "   Old: " + STR( nNo_Of_Old_Fields, 3 ) )
   else
      show( "Field Counts Match at " + STR( nNo_Of_Old_Fields, 3 ) )
   endif

   show( '' )

   IF nNo_Of_New_Fields <> nNo_Of_Old_Fields
      @ PROW() + 2, 5 SAY "File " + aOld_Just_File_Names[ nFiles_Counter, 1 ] + ;
         " has an incorrect number of fields."

      lPrint_First_Occurrence := .F.
      nNew_Fields_Counter := nOld_Fields_Counter := 0

      FOR nNew_Fields_Counter = 1 TO nNo_Of_New_Fields

         cCheck_Field := aNew_File_Structure_List[ nNew_Fields_Counter, 1 ]
         lGot_First_Occurrence := .F.

         FOR nOld_Fields_Counter = 1 TO nNo_Of_Old_Fields
            IF aOld_File_Structure_List[ nOld_Fields_Counter, 1 ] == ;
               cCheck_Field
               nOld_Fields_Counter := nNo_Of_Old_Fields
               lGot_First_Occurrence := .T.
            ENDIF
         NEXT

         IF !lGot_First_Occurrence
            IF !lPrint_First_Occurrence
               lPrint_First_Occurrence := .T.
               @ PROW() + 1, 7 SAY "New"
            ENDIF
            @ PROW() + 1, 10 SAY ALLTRIM( aNew_File_Structure_List[ nNew_Fields_Counter, 1 ] )
            @ PROW()    , 22 SAY ALLTRIM( aNew_File_Structure_List[ nNew_Fields_Counter, 2 ] )
            @ PROW()    , 25 SAY ALLTRIM( STR( aNew_File_Structure_List[ nNew_Fields_Counter, 3 ] ) )
            @ PROW()    , 30 SAY ALLTRIM( STR( aNew_File_Structure_List[ nNew_Fields_Counter, 4 ] ) )
         ENDIF

      NEXT

      lPrint_First_Occurrence := .F.
      nNew_Fields_Counter := nOld_Fields_Counter := 0

      FOR nOld_Fields_Counter = 1 TO nNo_Of_Old_Fields

         cCheck_Old_Field := aOld_File_Structure_List[nOld_Fields_Counter, 1]
         lGot_First_Occurrence := .F.

         FOR nNew_Fields_Counter = 1 TO nNo_Of_New_Fields
            IF aNew_File_Structure_List[ nNew_Fields_Counter, 1 ] == ;
               cCheck_Old_Field
               nNew_Fields_Counter := nNo_Of_New_Fields
               lGot_First_Occurrence := .T.
            ENDIF
         NEXT

         IF !lGot_First_Occurrence
            IF !lPrint_First_Occurrence
               lPrint_First_Occurrence := .T.
               @ PROW() + 1, 7 SAY "Old"
            ENDIF
            @ PROW() + 1, 10 SAY ALLTRIM( aOld_File_Structure_List[nOld_Fields_Counter, 1] )
            @ PROW()    , 22 SAY ALLTRIM( aOld_File_Structure_List[ nOld_Fields_Counter, 2 ] )
            @ PROW()    , 25 SAY ALLTRIM( STR( aOld_File_Structure_List[ nOld_Fields_Counter, 3 ] ) )
            @ PROW()    , 30 SAY ALLTRIM( STR( aOld_File_Structure_List[ nOld_Fields_Counter, 4 ] ) )
         ENDIF

      NEXT
   ENDIF

   /*
   if you get this far, check for structural differences this checks
   ordinal position, field name, type, length and decimal any
   differences will cause the check to end on the current file.
   */

   IF nNo_Of_New_Fields <> nNo_Of_Old_Fields
      @ PROW() + 2, 5 SAY "Can't check structure details for " + ;
         ALLTRIM( aOld_Just_File_Names[ nFiles_Counter, 1 ] ) + ;
         " because of previous "
      @ PROW() + 1, 5 SAY "differences found in the number of fields!"
   ELSE
      First_In_File := .T.
      // BEGIN SEQUENCE
      FOR nNew_Fields_Counter = 1 TO nNo_Of_New_Fields
         cCheck_Field := aNew_File_Structure_List[ nNew_Fields_Counter, 1 ]

         FOR nOld_Fields_Counter = 1 TO nNo_Of_Old_Fields
            IF aOld_File_Structure_List[ nOld_Fields_Counter, 1 ] == ;
               cCheck_Field

               lGot_A_Difference := .F.

               FOR nStructure_Counter = 1 to 4
                  IF aNew_File_Structure_List[ nNew_Fields_Counter, nStructure_Counter ] <> ;
                     aOld_File_Structure_List[ nOld_Fields_Counter, nStructure_Counter ]
                     lGot_A_Difference := .T.
                     nStructure_Counter := 4
                  ENDIF
               NEXT

               IF lGot_A_Difference
                  IF First_In_File
                     @ PROW() + 2, 5 SAY  "File " + aOld_Just_File_Names[ nFiles_Counter,1 ] + " has a structure change"
                     @ PROW() + 1, 10 SAY "New"
                     First_In_File := .F.
                  ELSE
                     @ PROW() + 2, 10 SAY "New"
                  ENDIF
                  @ PROW(), 15 SAY ALLTRIM( aNew_File_Structure_List[ nNew_Fields_Counter, 1 ] )
                  @ PROW(), 27 SAY ALLTRIM( aNew_File_Structure_List[ nNew_Fields_Counter, 2 ] )
                  @ PROW(), 30 SAY ALLTRIM( STR( aNew_File_Structure_List[ nNew_Fields_Counter, 3 ] ) )
                  @ PROW(), 35 SAY ALLTRIM( STR( aNew_File_Structure_List[ nNew_Fields_Counter, 4 ] ) )
                  @ PROW() + 1, 10 SAY "Old"
                  @ PROW(), 15 SAY ALLTRIM( aOld_File_Structure_List[ nOld_Fields_Counter, 1 ] )
                  @ PROW(), 27 SAY ALLTRIM( aOld_File_Structure_List[ nOld_Fields_Counter, 2 ] )
                  @ PROW(), 30 SAY ALLTRIM( STR( aOld_File_Structure_List[ nOld_Fields_Counter, 3 ] ) )
                  @ PROW(), 35 SAY ALLTRIM( STR( aOld_File_Structure_List[ nOld_Fields_Counter, 4 ] ) )
               ENDIF

               nOld_Fields_Counter := nNo_Of_Old_Fields

            ENDIF
         NEXT
      NEXT
      // END SEQUENCE
   ENDIF
NEXT

DBCLOSEALL()

/*
This loop checks to see if you have any
new files that the client does not have
*/

nNew_Files_Counter := nOld_Files_Counter := 0
aNew_Just_File_Names := DIRECTORY( ALLTRIM( cNew_Directory ) + "*.DBF" )

FOR nNew_Files_Counter := 1 TO LEN( aNew_Just_File_Names )

   cCheck_File := aNew_Just_File_Names[nNew_Files_Counter, 1 ]
   lGot_First_Occurrence := .F.

   FOR nOld_Files_Counter := 1 TO LEN( aOld_Just_File_Names )
      IF aOld_Just_File_Names[nOld_Files_Counter, 1 ] == ;
         cCheck_File
         nOld_Files_Counter := LEN( aOld_Just_File_Names )
         lGot_First_Occurrence := .T.
      ENDIF
   NEXT

   IF !lGot_First_Occurrence
      @ PROW() + 2, 5 SAY cCheck_File + ;
                            " is a new file that needs to be added."
   ENDIF

NEXT

EJECT

RETURN( NIL )

/*-------------------------------------------------------------------*/

FUNCTION FILLOLD( cNew_Directory, cOld_Directory )

LOCAL iValid:=.F., NulDev

IIF( RIGHT( ALLTRIM( cNew_Directory), 1)=="\", NulDev:="Nul", NulDev:="\Nul" )

IF ! EMPTY( cNew_Directory )
   IF FILE( ALLTRIM(cNew_Directory)+NulDev )
      IF EMPTY( cOld_Directory )
         cOld_Directory := cNew_Directory
      ENDIF
      iValid := HasDataFiles( cNew_Directory )
   ELSE
      TONE( 392, 3 )
      pop_msg( "Directory does not exist!" )
   ENDIF
ENDIF

RETURN iValid

/*-------------------------------------------------------------------*/

FUNCTION readypr()

* This warns to ready the printer

local Con

DO WHILE ! ISPRINTER() .and. Con <> "Q"

   TONE( 392, 3 )
   pop_msg( { "Printer is not ready. Please check!!!", "Press Q to Abort Process" } )
   con := lastkey()

ENDDO

RETURN( Con )

/*-------------------------------------------------------------------*/

FUNCTION ckfile( CKFILE, CKEXT )

LOCAL RET_VAL := .t.

IF FILE( ALLTRIM( CKFILE ) + "." + ALLTRIM( CKEXT ) )
   TONE( 392, 3 )
   POP_MSG( { "File already exists... ", "Enter another name! " } )
   RET_VAL := .F.
ENDIF

RETURN( RET_VAL )

/*-------------------------------------------------------------------*/

FUNCTION GETOutputType()

LOCAL GETLIST := {}

local GETOUTSCR := SAVESCREEN( 9, 16, 17, 64 )
local GTOUTCOLR := SETCOLOR()
LOCAL OUTFILE  := SPACE(8)
LOCAL OUTEXT   := SPACE(3)

SHADELITE ( 9, 16, 16, 62 )

setcolor( 'W+/BG+' )

dispbox( 9, 16, 16, 62, "ͻȺ " )

@ 10, 27 SAY "Output to :   Screen"
@ 11, 41 SAY "Printer"
@ 12, 41 SAY "File"
@ 13, 25 SAY "Output File :"
@ 14, 25 SAY "  Extension :"
@ 15, 25 SAY "  Laser/Dot :"

SETCOLOR( 'GR+/BG+' )

@ 10, 41 SAY "S"
@ 11, 41 SAY "P"
@ 12, 41 SAY "F"
@ 15, 27 SAY "L"
@ 15, 33 SAY "D"

OutputType := "S"

setcolor( 'GB+/B, GB+/B' )

@ 10, 39 GET OutputType PICTURE "@! A" VALID OutputType $ "SPF"
READ

do case
case OutputType == "P"
   OutputFile := MAKEOUTFIL()
case OutputType == "F"
   @ 13, 39 GET OUTFILE PICT "@! XXXXXXXX"
   @ 14, 39 GET OUTEXT  PICT "@! XXX" VALID CKFILE( OUTFILE, OUTEXT )
   read
   OutputFile := alltrim( outfile ) + '.' + alltrim( outext )
otherwise
   OutputFile := MAKEOUTFIL()
ENDcase

SETCOLOR( GTOUTCOLR )
RESTSCREEN( 9, 16, 17, 64, GETOUTSCR )

RETURN( OutputType )

/*-------------------------------------------------------------------*/

FUNCTION MAKEOUTFIL()

local DD := SUBSTR( DTOC( DATE() ), 4, 2 )
local HH := SUBSTR( TIME(), 1, 2)
local MM := SUBSTR( TIME(), 4, 2)
local SS := SUBSTR( TIME(), 7, 2)

RETURN( DD + HH + MM + SS + ".$$$" )  // $ is the universal temp indicator

/*-------------------------------------------------------------------*/

FUNCTION fsprint()

* Procedure to print temp file to printer

SET CONSOLE OFF
SET PRINT ON
SET PRINTER TO LPT1

COPY FILE ( OutputFile ) TO LPT1

SET PRINTER TO
SET PRINT OFF
SET CONSOLE ON

RETURN( NIL )

/*-------------------------------------------------------------------*/

FUNCTION shadelite( tp, lft, btt, rght )

local SHDLTCOLR := SETCOLOR()
local prvscr := SAVESCREEN( tp + 1, lft + 2, btt + 1, rght + 2 )

IF LEN(prvscr) > 2048
   prvscr := TRANSFORM(SUBSTR(prvscr,1,2048),;
      REPLICATE('X'+CHR(7),1024))+;
      TRANSFORM(SUBSTR(prvscr,2049),;
      REPLICATE('X'+CHR(7),(LEN(prvscr)-2048)/2))
ELSE
   prvscr := TRANSFORM(prvscr,;
      REPLICATE('X'+CHR(7),LEN(prvscr)/2))
ENDIF

RESTSCREEN(tp+1,lft+2,btt+1,rght+2,prvscr)

SETCOLOR( SHDLTCOLR )

RETURN( NIL )

/*-------------------------------------------------------------------*/

FUNCTION GoodOldDir( cDir )

LOCAL iValid:=.F., NulDev

cDir := alltrim( cDir )

  IIF( RIGHT( cDir, 1)=="\", NulDev:="Nul", NulDev:="\Nul"  )

  IF ! EMPTY( cDir )
   IF FILE( cDir + NulDev )
     iValid := HasDataFiles( cDir )
   ELSE
    TONE( 392, 3 )
    pop_msg( "Directory " + cDir + " does not exist!" )
   ENDIF
  ENDIF

RETURN iValid

/*-------------------------------------------------------------------*/

FUNCTION HasDataFiles( cDir )

local iValid := .T.
local FileSpec

cDir := alltrim( cDir )

FileSpec := IIF( RIGHT( cDir, 1 ) == "\", "*.DBF", "\*.DBF"  )

// Assumed Drive Only so get current directory
IF LEN( cDir ) == 2
   cDir := SUBSTR( cDir, 1, 2 ) + "\" + RTRIM( CURDIR( cDir ) )
ENDIF

IF ! FILE( cDir + FileSpec )
   iValid:=.F.
   TONE( 392, 3 )
   pop_msg( "Directory has no data files!" )
ENDIF

RETURN iValid

/*-------------------------------------------------------------------*/

function show( cText )

SET DEVICE TO SCREEN

scroll( 11, 1, 20, 78, 1 )

@ 20, 2 say left( cText, 78 )

SET DEVICE TO PRINT

return NIL
